"""
This module implements qiskit's kak as a native tool plugin to QFAST.
"""

import qiskit
import numpy as np


def get_native_block_size():
    """
    The maximum size of a unitary matrix (in qubits) that can be
    decomposed with this module.

    Returns:
        (int): The qubit count this module can handle.
    """

    return 2


def synthesize ( utry ):
    """
    Synthesis function with QISKit's KAK implementation.

    Args:
        utry (np.ndarray): The unitary matrix to synthesize.

    Returns
        qasm (str): The synthesized QASM output.
    """

    if not isinstance( utry, np.ndarray ):
        raise TypeError( "utry must be a np.ndarray." )

    if len( utry.shape ) != 2:
        raise TypeError( "utry must be a matrix." )

    if utry.shape[0] != 2 ** get_native_block_size():
        raise ValueError( "utry has incorrect dimensions." )

    if utry.shape[1] != 2 ** get_native_block_size():
        raise ValueError( "utry has incorrect dimensions." )

    if ( not np.allclose( utry.conj().T @ utry, np.identity( len( utry ) ),
                          rtol = 0, atol = 1e-14 )
         or
         not np.allclose( utry @ utry.conj().T, np.identity( len( utry ) ),
                          rtol = 0, atol = 1e-14 ) ):
        raise ValueError( "utry must be a unitary matrix." )

    circ = qiskit.QuantumCircuit( get_native_block_size() )
    circ.unitary( utry, [ 1, 0 ] )
    circ = qiskit.compiler.transpile( circ, basis_gates = ['u3', 'cx'],
                                      optimization_level = 3 )
    return circ.qasm()





"""
This module implements a Pauli Operator Library.

The QFAST Compiler relies on pauli operators tremendously.
All the necessary objects and functions are implemented in this library.
"""

import scipy
import numpy      as np
import itertools  as it
import tensorflow as tf

# The Pauli Matrices
X = np.array( [ [ 0, 1 ],
                [ 1, 0 ] ], dtype = np.complex128 )

Y = np.array( [ [ 0, -1j ],
                [ 1j,  0 ] ], dtype = np.complex128 )

Z = np.array( [ [ 1,  0 ],
                [ 0, -1 ] ], dtype = np.complex128 )

I = np.array( [ [ 1, 0 ],
                [ 0, 1 ] ], dtype = np.complex128 )


# Pauli Cache
norder_paulis_map = [ [ I ], [ I, X, Y, Z ] ]
norder_paulis_tensor_map = [ [ tf.constant( I ) ], [ tf.constant( I ),
                                                     tf.constant( X ),
                                                     tf.constant( Y ),
                                                     tf.constant( Z ) ] ]


def reset_tensor_cache():
    norder_paulis_tensor_map.clear()
    norder_paulis_tensor_map.append( [ tf.constant( I ) ] )
    norder_paulis_tensor_map.append( [ tf.constant( I ),
                                       tf.constant( X ),
                                       tf.constant( Y ),
                                       tf.constant( Z ) ] )


def get_norder_paulis ( n ):
    """
    Recursively constructs the nth-order tensor product of the Pauli group.

    Args:
        n (int): Power of the tensor product of the Pauli group.

    Returns:
        (List[np.ndarray]): nth-order Pauli matrices
    """

    if n < 0:
        raise ValueError( "n must be a dimension greater than or equal to 0." )

    if len( norder_paulis_map ) > n:
        return norder_paulis_map[n]

    norder_paulis = []
    for pauli_n_1, pauli_1 in it.product( get_norder_paulis( n - 1 ),
                                          get_norder_paulis(1) ):
        norder_paulis.append( np.kron( pauli_n_1, pauli_1 ) )

    norder_paulis_map.append( norder_paulis )

    return norder_paulis_map[n]


def get_norder_paulis_tensor ( n ):
    """
    Retrieve TensorFlow versions of the Paulis from the cache.

    Args:
        n (int): Power of the tensor product of the Pauli group.

    Returns:
        (List[tf.tensor]): nth-order Pauli matrices
    """

    if n < 0:
        raise ValueError( "n must be a dimension greater than or equal to 0." )

    if len( norder_paulis_tensor_map ) > n:
        return norder_paulis_tensor_map[n]

    get_norder_paulis_tensor( n - 1 )

    paulis_tensor = [ tf.constant( p ) for p in get_norder_paulis( n ) ]
    norder_paulis_tensor_map.append( paulis_tensor )

    return norder_paulis_tensor_map[n]


def get_pauli_n_qubit_projection ( n, q_list ):
    """
    Returns the nth-order Pauli matrices that act only on qubits in q_list.

    Args:
        n (int): Power of the tensor product of the Pauli group

        q_list (Tuple[int]): Qubit indices

    Returns:
        pauli_n_qubit (List[np.array]): nth-order Pauli matrices acting
                                        only on qubits in q_list.
    """

    if any( [ q < 0 or q >= n for q in q_list ] ):
        raise ValueError( "Qubit indices must be in [0, n).")

    if len( q_list ) != len( set( q_list ) ):
        raise ValueError( "Qubit indices cannot be duplicates." )

    if len( q_list ) == 0:
        raise ValueError( "Need atleast one qubit index." )

    paulis = get_norder_paulis( n )

    # Nth Order Pauli Matrices can be thought of base 4 number
    # I = 0, X = 1, Y = 2, Z = 3
    # XXY = 1 * 4^2 + 1 & 4^1 + 2 * 4^0 = 22 (base 10)
    # This gives the idx of XXY in paulis
    # Note we read qubit index from the left,
    # so X in XII corresponds to q = 0
    pauli_n_qubit = []
    for ps in it.product( [ 0, 1, 2, 3 ], repeat = len( q_list ) ):
        idx = 0
        for p, q in zip( ps, q_list ):
            idx += p * ( 4 ** ( n - q - 1 ) )
        pauli_n_qubit.append( paulis[ idx ] )

    return pauli_n_qubit


def get_pauli_tensor_n_qubit_projection ( n, q_list ):
    """
    Returns the nth-order Pauli matrices that act only on qubits in q_list
    in tensor type.

    Args:
        n (int): Power of the tensor product of the Pauli group

        q_list (Tuple[int]): Qubit indices

    Returns:
        pauli_n_qubit (List[tf.tensor]): nth-order Pauli matrices acting
                                         only on qubits in q_list.
    """

    if any( [ q < 0 or q >= n for q in q_list ] ):
        raise ValueError( "Qubit indices must be in [0, n).")

    if len( q_list ) != len( set( q_list ) ):
        raise ValueError( "Qubit indices cannot be duplicates." )

    if len( q_list ) == 0:
        raise ValueError( "Need atleast one qubit index." )

    paulis = get_norder_paulis_tensor( n )

    # Nth Order Pauli Matrices can be thought of base 4 number
    # I = 0, X = 1, Y = 2, Z = 3
    # XXY = 1 * 4^2 + 1 & 4^1 + 2 * 4^0 = 22 (base 10)
    # This gives the idx of XXY in paulis
    # Note we read qubit index from the left,
    # so X in XII corresponds to q = 0
    pauli_n_qubit = []
    for ps in it.product( [ 0, 1, 2, 3 ], repeat = len( q_list ) ):
        idx = 0
        for p, q in zip( ps, q_list ):
            idx += p * ( 4 ** ( n - q - 1 ) )
        pauli_n_qubit.append( paulis[ idx ] )

    return pauli_n_qubit


def pauli_dot_product ( alpha, sigma ):
    """
    Computes the dot product of alpha and sigma.

    Args:
        alpha (List[float]): The alpha coefficients

        sigma (List[np.ndarray]): The Pauli Matrices

    Returns:
        (np.ndarray): Hermitian matrix computed from the dot
                      product of alpha and sigma.
    """

    if len( alpha ) != len( sigma ):
        raise ValueError( "Length of alpha and sigma must be the same." )

    return np.sum( [ a*s for a, s in zip( alpha, sigma ) ], 0 )


def get_unitary_from_pauli_coefs ( pauli_coefs ):
    """
    Convert a pauli expansion to a unitary matrix.

    Args:
        pauli_coefs (List[float]): Coefficient of Pauli matrices

    Returns:
        (np.ndarray): Unitary Matrix computed from the Pauli coefficients
    """

    num_qubits = int( np.log2( len( pauli_coefs ) ) / 2 )
    sigma = get_norder_paulis( num_qubits )

    if len( sigma ) != len( pauli_coefs ):
        raise ValueError( "Invalid pauli_coefs list." )

    H = pauli_dot_product( pauli_coefs, sigma )
    return scipy.linalg.expm( 1j * H )


def unitary_log_no_i ( U ):
    """
    Solves for H in U = e^{iH}

    Args:
        U (np.ndarray): The unitary to decompose

    Returns:
        H (np.ndarray): e^{iH} = U
    """

    if ( not np.allclose( U.conj().T @ U, np.identity( len( U ) ),
                          rtol = 0, atol = 1e-14 )
         or
         not np.allclose( U @ U.conj().T, np.identity( len( U ) ),
                          rtol = 0, atol = 1e-14 ) ):
        raise ValueError( "U must be a unitary matrix." )

    T, Z = scipy.linalg.schur( U, output = 'complex' )
    assert( np.allclose( Z @ T @ Z.conj().T, U ) )
    H = np.diag( np.log( np.diagonal( T ) ) )
    H = -1j*H
    H = np.matmul( np.matmul( Z, H ), Z.conj().T )
    return H


def unitary_log_no_i_eig ( U ):
    """
    Solves for H in U = e^{iH}

    Args:
        U (np.ndarray): The unitary to decompose

    Returns:
        H (np.ndarray): e^{iH} = U
    """

    if ( not np.allclose( U.conj().T @ U, np.identity( len( U ) ),
                          rtol = 0, atol = 1e-14 )
         or
         not np.allclose( U @ U.conj().T, np.identity( len( U ) ),
                          rtol = 0, atol = 1e-14 ) ):
        raise ValueError( "U must be a unitary matrix." )

    T, Z = scipy.linalg.eig( U )
    angles = -1j * np.log( T )
    H = np.diag( angles )
    H = np.matmul( np.matmul( Z, H ), Z.conj().T )
    return H


def pauli_expansion ( H ):
    """
    Computes a Pauli expansion of the hermitian matrix H.

    Args:
        H (np.ndarray): The hermitian matrix

    Returns:
        X (list of floats): The coefficients of a Pauli expansion for H,
                            i.e., X dot Sigma = H where Sigma is
                            Pauli matrices of same size of H
    """

    if not np.allclose( H, H.conj().T, rtol = 0, atol = 1e-14 ):
        raise ValueError( "H must be hermitian." )

    # Change basis of H to Pauli Basis (solve for coefficients -> X)
    n = int( np.log2( len( H ) ) )
    paulis = get_norder_paulis( n )
    flatten_paulis = [ np.reshape( pauli, 4 ** n ) for pauli in paulis ]
    flatten_H      = np.reshape( H, 4 ** n )
    A = np.stack( flatten_paulis, axis = -1 )
    X = np.real( np.matmul( np.linalg.inv( A ), flatten_H ) )
    return X



from .fixedgate     import FixedGate
from .genericgate   import GenericGate
from .block         import Block
from .decomposition import get_decomposition_size
from .decomposition import fixed_depth_exploration
from .decomposition import exploration
from .decomposition import refinement
from .decomposition import convert_to_block_list
from .decomposition import decomposition
from .circuit       import Circuit
from .metrics       import hilbert_schmidt_distance
from .pauli         import get_norder_paulis
from .pauli         import get_norder_paulis_tensor
from .pauli         import get_pauli_n_qubit_projection
from .pauli         import get_unitary_from_pauli_coefs
from .pauli         import get_pauli_tensor_n_qubit_projection
from .pauli         import pauli_dot_product
from .pauli         import pauli_expansion
from .pauli         import unitary_log_no_i, unitary_log_no_i_eig
from .pauli         import reset_tensor_cache, I, X, Y, Z
from .recombination import recombination
from .instantiation import list_native_tools
from .instantiation import get_native_tool
from .instantiation import instantiation
from .locationmodel import greedy_max_cut
from .locationmodel import lexicographical_cut
from .locationmodel import LocationModel





"""
This module implements the functions for the recombination phase.
"""

from qiskit import *

def recombination ( qasm_list, loc_fixed ):
    """
    Recombines the circuits in qasm_list into one circuit.

    Args:
        qasm_list (List[str]): The list of circuits given by QASM

        loc_fixed (List[Tuple[int]]): The circuit locations

    Returns:
        (str): The final circuit's QASM
    """

    if not isinstance( qasm_list, list ):
        raise TypeError( "qasm_list must be a list." )

    if not all( isinstance( qasm, str ) for qasm in qasm_list ):
        raise TypeError( "qasm_list must contain QASM strings." )

    if not isinstance( loc_fixed, list ):
        raise TypeError( "loc_fixed must be a list." )

    if ( not all( isinstance( loc, tuple ) for loc in loc_fixed )
         or not all( len( loc ) == len( set( loc ) ) for loc in loc_fixed )
         or not all( isinstance( q, int ) for loc in loc_fixed for q in loc ) ):
        raise TypeError( "loc_fixed must contain valid locations." )

    # Calculate Output Circuit Size
    max_qubit = 0

    for loc in loc_fixed:
        for qub in loc:
            max_qubit = max( qub, max_qubit )

    # Convert to circuits
    circs = []
    for qasm in qasm_list:
        try:
            circs.append( QuantumCircuit.from_qasm_str( qasm ) )
        except:
            raise ValueError( "Invalid QASM string: %s" % qasm )

    if all( len( circ.qubits ) != len( loc )
            for circ, loc in zip( circs, loc_fixed ) ):
        raise ValueError( "Location and QASM qubit counts don't match." )

    # Join into One Circuit
    out_circ = QuantumCircuit( max_qubit + 1 )

    for circ, loc in zip( circs, loc_fixed ):
        for gate in circ.data:
            if gate[0].name == 'cx':
                out_circ.cx( loc[ gate[1][0].index ], loc[ gate[1][1].index ] )
            elif gate[0].name == 'u1':
                out_circ.u1( *gate[0].params, loc[ gate[1][0].index ] )
            elif gate[0].name == 'u2':
                out_circ.u2( *gate[0].params, loc[ gate[1][0].index ] )
            elif gate[0].name == 'u3':
                out_circ.u3( *gate[0].params, loc[ gate[1][0].index ] )
            elif gate[0].name == 'rx':
                out_circ.rx( *gate[0].params, loc[ gate[1][0].index ] )
            elif gate[0].name == 'ry':
                out_circ.ry( *gate[0].params, loc[ gate[1][0].index ] )
            elif gate[0].name == 'rz':
                out_circ.rz( *gate[0].params, loc[ gate[1][0].index ] )
            else:
                raise ValueError( "QASM must be in \'u1, u2, u3, cx\' basis." )

    out_circ = qiskit.compiler.transpile( out_circ, basis_gates = ['u3', 'cx'],
                                          optimization_level = 3 )
    return out_circ.qasm()



"""
This module implements the location model class.

The location model is responsible for mapping location indices
to locations. This is useful in gate parity calculations.
"""

import numpy as np
import itertools as it


def greedy_max_cut ( locations ):
    """
    Partitions the locations into two buckets by placing them as nodes
    in a graph, where edges connect two locations with at least one
    shared qubit, and greedily approximating a maximum cut.

    Args:
        locations (Set[Tuple[int]]): The location set

    Returns:
        buckets (List[List[Tuple[int]]]): The bucketed locations
    """

    # Create Graph as adjacency list, store cut info in nodes
    graph = { l:[] for l in locations }
    nodes = { l:{ "cut":0, "uncut":0 } for l in locations }
    edges = set()
    for n1, n2 in it.product( locations, repeat = 2 ):
        if n1 == n2:
            continue
        for q in n1:
            if q in n2 and ( n2, n1 ) not in edges:
                edges.add( ( n1, n2 ) )
                graph[ n1 ].append( n2 )
                graph[ n2 ].append( n1 )
                break

    # Start with empty cut
    for node, adj_node_list in graph.items():
        nodes[ node ][ "uncut" ] = len( adj_node_list )

    buckets = [ [], list( nodes.keys() ) ]

    while True:
        # Sort nodes by possible value gain
        sorted_nodes = sorted( nodes.items(),
                               key = lambda x: x[1]["uncut"] - x[1]["cut"],
                               reverse = True )

        nd = sorted_nodes[0]

        # If no more possible value gain, stop
        if nd[1]["uncut"] - nd[1]["cut"] <= 0:
            break

        # Swap nd across the cut
        buckets[0].append( nd[0] )
        buckets[1].remove( nd[0] )
        temp = nodes[ nd[0] ][ 'cut' ]
        nodes[ nd[0] ][ 'cut' ] = nodes[ nd[0] ][ 'uncut' ]
        nodes[ nd[0] ][ 'uncut' ] = temp
        for neighbor in graph[ nd[0] ]:
            nodes[ neighbor ][ 'cut' ] += 1
            nodes[ neighbor ][ 'uncut' ] -= 1

    return buckets


def lexicographical_cut ( locations ):
    """
    Partitions the locations into two buckets by cutting in half.

    Args:
        locations (Set[Tuple[int]]): The location set

    Returns:
        buckets (List[List[Tuple[int]]]): The bucketed locations
    """

    locations = list( locations )
    n = len( locations )
    return [ locations[:n//2], locations[n//2:] ]


class LocationModel():
    """The LocationModel class."""

    def __init__ ( self, num_qubits, gate_size, cgraph = None,
                   bucketing_alg = None ):
        """
        Location model construction.

        Args:
            num_qubits (int): Number of qubits in the target unitary

            gate_size (int): Number of qubits in each gate

            cgraph (Set[Tuple[int]]): Coupling graph, defaults to
                                      fully connected hardware topology

            bucketing_alg (Set[Tuple[int]] -> List[List[Tuple[int]]]):
                Function that converts a set of locations into buckets
                of locations. Defaults to lexicographical cut.
        """

        if num_qubits <= 0:
            raise ValueError( "Invalid qubit count: %d." % num_qubits )

        if gate_size > num_qubits or gate_size <= 0:
            raise ValueError( "Invalid gate size: %d." % gate_size )

        if cgraph is None:
            cgraph = set( it.combinations( range( num_qubits ), 2 ) )

        if len( cgraph ) != len( set( cgraph ) ):
            raise ValueError( "cgraph contains duplicates: %s", str( cgraph ) )

        if ( not all( [ ( 0 <= q < num_qubits ) for l in cgraph for q in l ] )
             or not all( [ len( l ) == len( set( l ) ) for l in cgraph ] )
             or not all( [ len( l ) == 2 for l in cgraph ] ) ):
            raise ValueError( "Invalid cgraph supplied: %s.", str( cgraph ) )

        if bucketing_alg is None:
            bucketing_alg = lexicographical_cut

        self.num_qubits = num_qubits
        self.gate_size = gate_size
        self.cgraph = cgraph
        self.bucketing_alg = bucketing_alg

        # TODO Topology Aware: change locations to be a function of cgraph
        self.locations = it.combinations( range( num_qubits ), gate_size )
        self.locations = set( self.locations )

        self.buckets = bucketing_alg( self.locations )
        self.num_buckets = len( self.buckets )

        # Check if bucketing algorithm returned valid buckets.
        bucket_sum = set()
        element_sum = 0
        for b in self.buckets:
            bucket_sum.update( b )
            element_sum += len( b )
        if bucket_sum != self.locations or element_sum != len( self.locations ):
            raise ValueError( "Invalid bucketing algorithm, not a partition." )

    def fix_locations ( self, loc_vals ):
        """
        Fixes the locations; used when converting from generic gates to
        fixed gates.

        Args:
            loc_vals (List[List[float]]): Gate unfixed location values

        Returns:
            (List[Tuple[int]]): Fixed locations
        """

        loc_fixed = []

        for i, loc_val in enumerate( loc_vals ):
            loc_idx = np.argmax( loc_val )
            parity = (i + 1) % self.num_buckets
            loc_fixed.append( self.buckets[ parity ][ loc_idx ] )

        return loc_fixed



import argparse

import numpy as np
from qiskit import *


def flip_circ ( circ ):
    num_qubits = len( circ.qubits )

    if num_qubits == 1:
        raise ValueError( "Cannot flip 1-qubit circuit." )

    circ_flip = QuantumCircuit( num_qubits )
    for gate in circ.data:
        if gate[0].name == 'cx':
            circ_flip.cx( num_qubits - 1 - gate[1][0].index, num_qubits - 1 - gate[1][1].index )
        elif gate[0].name == 'u3':
            circ_flip.u3( *gate[0].params, num_qubits - 1 - gate[1][0].index )
        elif gate[0].name == 'u2':
            circ_flip.u2( *gate[0].params, num_qubits - 1 - gate[1][0].index )
        elif gate[0].name == 'u1':
            circ_flip.u1( *gate[0].params, num_qubits - 1 - gate[1][0].index )
        elif gate[0].name == 'barrier':
            continue
        else:
            raise ValueError( "Cannot flip circuit not in {U1, U2, U3, CX} basis." )
    return circ_flip


def calc_unitary ( circ ):
    backend = BasicAer.get_backend( 'unitary_simulator' )
    return qiskit.execute( circ, backend ).result().get_unitary()


def test_unitary ( circ, utry ):
    # Test unitary shape
    num_qubits = len( circ.qubits )
    if utry.shape != (2**num_qubits, 2**num_qubits):
        raise RuntimeError( "Unitary failed verification test: wrong shape." )

    # Test unitary matrix
    backend = BasicAer.get_backend( 'statevector_simulator' )

    for idx in range( 2**num_qubits ):
        # For all standard basis vectors (Have to reverse for qiskit)
        base_vec = np.array( [ int(x == idx) for x in range( 2**num_qubits ) ] )
        revd_idx = int( ('{0:0%db}' % num_qubits).format(idx)[::-1], 2 )
        revd_vec = np.array( [ int(x == revd_idx) for x in range( 2**num_qubits ) ] )

        # Calculate correct probability distribution
        boptions = { 'initial_statevector': revd_vec }
        verf_job = qiskit.execute( circ, backend, backend_options = boptions )
        verf_vec = verf_job.result().get_statevector()

        verf_dist = {}
        for x in range( 2**num_qubits ):
            bit_str = ('{0:0%db}' % num_qubits).format(x)
            bit_str = bit_str[::-1]  # reverse( bit_str )
            bit_prb = np.abs( verf_vec[x] ) ** 2
            verf_dist[ bit_str ] = bit_prb

        # Calculate test probability distribution
        test_vec = utry @ np.reshape( base_vec, ( len( base_vec ), 1 ) )
        test_vec = np.squeeze( test_vec )

        test_dist = {}
        for x in range( 2**num_qubits ):
            bit_str = ('{0:0%db}' % num_qubits).format(x)
            bit_prb = np.abs( test_vec[x] ) ** 2
            test_dist[ bit_str ] = bit_prb

        # Test equality of distributions
        for x in range( 2**num_qubits ):
            bit_str = ('{0:0%db}' % num_qubits).format(x)
            if not np.allclose( verf_dist[ bit_str ], test_dist[ bit_str ] ):
                raise RuntimeError( "Unitary failed verification test: incorrect matrix" )


def hilbert_schmidt_distance ( X, Y ):
    """
    Calculates a distance based on the Hilbert Schmidt inner product.

    Args:
        X: First Operator
        Y: Second Operator

    Returns:
        Error value between X and Y
    """

    if X.shape != Y.shape:
        raise ValueError( "X and Y must have same shape." )

    mat = np.matmul( np.transpose( np.conj( X ) ), Y )
    num = np.abs( np.trace( mat ) ) ** 2
    dem = mat.shape[0] ** 2
    return np.sqrt( 1 - ( num / dem ) )


if __name__ == "__main__":
    description_info = "Convert QASM to a unitary."

    parser = argparse.ArgumentParser( description = description_info )

    parser.add_argument( "qasm_file", type = str,
                         help = "QASM input file" )

    parser.add_argument( "unitary_file", type = str,
                         help = "Unitary output file" )

    args = parser.parse_args()

    circ = QuantumCircuit.from_qasm_file( args.qasm_file )
    circ.remove_final_measurements()
    utry = calc_unitary( flip_circ( circ ) )
    test_unitary( circ, utry )

    target = np.loadtxt( args.unitary_file, dtype = np.complex128 )
    print( hilbert_schmidt_distance( utry, target ) )





import os
import sys
import logging
import numpy    as np
import argparse as ap

from .circuit import Circuit
from .instantiation import *
from .recombination import recombination


logger = logging.getLogger( "qfast" )


if __name__ == "__main__":
    description_info = "Synthesize a unitary matrix."

    parser = ap.ArgumentParser( description = description_info )

    parser.add_argument( "--decompose-only",
                         default = False,
                         action = "store_true",
                         help = "Perform decomposition" )

    parser.add_argument( "--instantiate-only",
                         default = False,
                         action = "store_true",
                         help = "Perform instantiation" )

    parser.add_argument( "--recombine-only",
                         default = False,
                         action = "store_true",
                         help = "Perform recombination" )

    parser.add_argument( "--unitary-file",
                         type = str,
                         default = None,
                         help = "Unitary file input" )

    parser.add_argument( "--unitary-dir",
                         type = str,
                         default = None,
                         help = "Directory of unitary files" )

    parser.add_argument( "--qasm-dir",
                         type = str,
                         default = None,
                         help = "Directory of qasm files" )

    parser.add_argument( "--qasm-file",
                         type = str,
                         default = None,
                         help = "Qasm file output" )

    parser.add_argument( "-n", "--native-tool",
                         type = str,
                         default = None,
                         choices = list_native_tools(),
                         help = "The tool to use during instantiation." )

    parser.add_argument( "-b", "--block-size",
                         type = int,
                         default = 2,
                         help = "The block size to decompose to." )

    parser.add_argument( "-s", "--start-depth", dest = "sd",
                         type = int,
                         default = 1,
                         help = "The initial depth to start exploring from." )

    parser.add_argument( "-d", "--depth-step", dest = "ds",
                         type = int,
                         default = 1,
                         help = ( "The number of gates to append every"
                                  "step of exploration." ) )

    parser.add_argument( "-e", "--exploration-distance", dest = "ed",
                         type = float,
                         default = 0.01,
                         help = "The distance tolerance for exploration." )

    parser.add_argument( "-l", "--exploration-learning-rate", dest = "el",
                         type = float,
                         default = 0.01,
                         help = "The learning rate for exploration." )

    parser.add_argument( "-r", "--refinement-distance", dest = "rd",
                         type = float,
                         default = 1e-7,
                         help = "The distance tolerance for refinement." )

    parser.add_argument( "-j", "--refinement-learning-rate", dest = "rl",
                         type = float,
                         default = 1e-6,
                         help = "The learning rate for refinement." )

    parser.add_argument( "-v", "--verbose",
                         default = 0,
                         action = "count",
                         help = "Verbose output" )

    args = parser.parse_args()

    # True if we have to do everything
    complete_qfast = not ( args.decompose_only or
                           args.instantiate_only or
                           args.recombine_only )

    # Can only have one "only" flag set
    if int( args.decompose_only ) + \
       int( args.instantiate_only ) + \
       int( args.recombine_only ) > 1:
        parser.error( "Can only have one stage-only flag set" )

    # If we need to do everything need unitary file and qasm file
    if complete_qfast:
        if args.unitary_file is None:
            parser.error( "No unitary input file specified"
                          ", add --unitary-file." )
        if args.qasm_file is None:
            parser.error( "No qasm output file specified"
                          ", add --qasm-file." )
        if args.native_tool is None:
            parser.error( "No native tool specified"
                          ", add --native-tool." )

    # If we are only doing decomposition need unitary file and dir
    if args.decompose_only:
        if args.unitary_file is None:
            parser.error( "No unitary input file specified"
                          ", add --unitary-file." )
        if args.unitary_dir is None:
            parser.error( "No unitary output directory specified"
                          ", add --unitary-dir." )

    # If we are only doing instantiation need unitary dir and qasm dir
    if args.instantiate_only:
        if args.unitary_dir is None:
            parser.error( "No unitary input directory specified"
                          ", add --unitary-dir." )
        if args.qasm_dir is None:
            parser.error( "No qasm output directory specified"
                          ", add --qasm-dir." )
        if args.native_tool is None:
            parser.error( "No native tool specified"
                          ", add --native-tool." )

    # If we are only doing recombination need qasm dir and qasm file
    if args.recombine_only:
        if args.qasm_dir is None:
            parser.error( "No qasm input directory specified"
                          ", add --qasm-dir." )
        if args.qasm_file is None:
            parser.error( "No qasm output file specified"
                          ", add --qasm-file." )

    # Logging Init
    handler = logging.StreamHandler( sys.stdout )

    if args.verbose >= 3:
        logger.setLevel( logging.DEBUG )
    elif args.verbose == 2:
        handler.addFilter( lambda m: m.getMessage()[:4] != "Loss" )
        logger.setLevel( logging.DEBUG )
    elif args.verbose == 1:
        logger.setLevel( logging.INFO )
    elif args.verbose == 0:
        logger.setLevel( logging.CRITICAL )
        logging.getLogger( "tensorflow" ).addFilter( lambda m: False )

    handler.setLevel( logging.DEBUG )
    logger.addHandler( handler )

    if args.decompose_only:
        target = np.loadtxt( args.unitary_file, dtype = np.complex128 )
        circ = Circuit( target )
        circ.hierarchically_decompose( args.block_size,
                                       start_depth = args.sd,
                                       depth_step = args.ds,
                                       exploration_distance = args.ed,
                                       exploration_learning_rate = args.el,
                                       refinement_distance = args.rd,
                                       refinement_learning_rate = args.rl )

        if not os.path.isdir( args.unitary_dir ):
            os.makedirs( args.unitary_dir )

        circ.dump_blocks( args.unitary_dir )

    elif args.instantiate_only:
        if not os.path.isdir( args.unitary_dir ):
            raise RuntimeError( "Unitary directory does not exist." )

        if not os.path.isdir( args.qasm_dir ):
            os.makedirs( args.qasm_dir )

        for file in os.listdir( args.unitary_dir ):
            utry = np.loadtxt( os.path.join( args.unitary_dir, file ),
                               dtype = np.complex128 )
            name = ".".join( file.split(".")[:-1] ) + ".qasm"
            qasm = instantiation( args.native_tool, utry )
            with open( os.path.join( args.qasm_dir, name ), 'w' ) as f:
                f.write( qasm )

    elif args.recombine_only:
        if not os.path.isdir( args.qasm_dir ):
            raise RuntimeError( "Qasm directory does not exist." )

        qasm_list_dict = {}
        loc_fixed_dict = {}

        for file in os.listdir( args.qasm_dir ):
            with open( os.path.join( args.qasm_dir, file ), "r" ) as f:
                qasm = f.read()
            name_list = file.replace( ".qasm", "" ).split("_")
            loc = tuple( int(x) for x in name_list[1:] )
            idx = int( name_list[0] )
            qasm_list_dict[ idx ] = qasm
            loc_fixed_dict[ idx ] = loc

        qasm_list = []
        loc_fixed = []

        for i in range( len( os.listdir( args.qasm_dir ) ) ):
            qasm_list.append( qasm_list_dict[ i ] )
            loc_fixed.append( loc_fixed_dict[ i ] )

        out_qasm = recombination( qasm_list, loc_fixed )

        with open( args.qasm_file, "w" ) as f:
            f.write( out_qasm )

    elif complete_qfast:
        target = np.loadtxt( args.unitary_file, dtype = np.complex128 )
        circ = Circuit( target )
        block_size = get_native_tool( args.native_tool ).get_native_block_size()
        circ.hierarchically_decompose( block_size,
                                       start_depth = args.sd,
                                       depth_step = args.ds,
                                       exploration_distance = args.ed,
                                       exploration_learning_rate = args.el,
                                       refinement_distance = args.rd,
                                       refinement_learning_rate = args.rl )

        qasm_list = [ instantiation( args.native_tool, block.utry )
                      for block in circ.blocks ]
        locations = circ.get_locations()

        out_qasm = recombination( qasm_list, locations )

        with open( args.qasm_file, "w" ) as f:
            f.write( out_qasm )



"""
This module implements the Block Class.

A block is a unitary operation applied to a set of qubits.
"""

import numpy as np

from .pauli import unitary_log_no_i, pauli_expansion


class Block():
    """
    The Block Class.
    """

    def __init__ ( self, utry, loc ):
        """
        Block Class Constructor

        Args:
            utry (np.ndarray): Unitary

            location (tuple[int]): Block location (set of qubits)
        """

        if not isinstance( utry, np.ndarray ):
            raise TypeError( "utry must be a np.ndarray." )

        if len( utry.shape ) != 2:
            raise TypeError( "utry must be a matrix." )

        if not isinstance( loc, tuple ):
            raise TypeError( "loc must be a location." )

        if not all( isinstance( q, int ) for q in loc ):
            raise TypeError( "loc must be a location." )

        if not len( loc ) == len( set( loc ) ):
            raise TypeError( "loc must be a valid location." )

        if 2 ** len( loc ) != utry.shape[0]:
            raise ValueError( "loc and utry have incompatible dimensions." )

        if 2 ** len( loc ) != utry.shape[1]:
            raise ValueError( "loc and utry have incompatible dimensions." )

        if not np.allclose( utry @ utry.conj().T, np.identity( len( utry ) ) ):
            raise ValueError( "utry must be a unitary matrix." )

        if not np.allclose( utry.conj().T @ utry, np.identity( len( utry ) ) ):
            raise ValueError( "utry must be a unitary matrix." )

        self.utry = utry
        self.loc  = loc
        self.num_qubits = len( loc )

    def get_utry ( self ):
        """
        Gets the block's unitary.

        Returns:
            (np.ndarray): Block's unitary
        """

        return self.utry

    def get_fun_vals ( self ):
        """
        Gets the block's function parameters.

        Returns:
            (List[float]): Block's function parameters
        """

        return pauli_expansion( unitary_log_no_i( self.utry ) )

    def get_location ( self ):
        """
        Gets the block's location.

        Returns:
            (Tuple[int]): Block's location
        """

        return self.loc

    def get_num_qubits ( self ):
        """
        Gets the block's size in qubits.

        Returns:
            (int): Block's size
        """

        return self.num_qubits

    def __str__ ( self ):
        """
        Gets the block's string representation.

        Returns:
            (str): Block's string representation
        """

        return str( self.loc ) + ":" + str( self.utry )

    def __repr__ ( self ):
        """
        Gets a simple block string representation.

        Returns:
            (str): Block's simple string representation
        """

        return str( self.loc )           \
               + ": [["                    \
               + str( self.utry[0][0] )   \
               + " ... "                  \
               + str( self.utry[-1][-1] ) \
               + "]]"



"""
This module implements unitary distance functions.
"""

import tensorflow as tf
import numpy      as np


def hilbert_schmidt_distance ( X, Y ):
    """
    Calculates a distance based on the Hilbert Schmidt inner product.

    Args:
        X: First Operator
        Y: Second Operator

    Returns:
        Either a result in numpy or a tensor in tensorflow.
    """

    if not isinstance( X, np.ndarray ) and not isinstance( X, tf.Tensor ):
        raise TypeError( "X must be either a np.ndarray or a tf.Tensor." )

    if not isinstance( Y, np.ndarray ) and not isinstance( Y, tf.Tensor ):
        raise TypeError( "Y must be either a np.ndarray or a tf.Tensor." )

    if X.shape != Y.shape:
        raise ValueError( "X and Y must have same shape." )

    if isinstance( X, np.ndarray ) and isinstance( Y, np.ndarray ):
        mat = np.matmul( np.transpose( np.conj( X ) ), Y )
        num = np.abs( np.trace( mat ) ) ** 2
        dem = mat.shape[0] ** 2

        quotient = num / dem
        if np.allclose( 1, quotient, rtol = 0, atol = 1e-15 ):
            quotient = 1

        return np.sqrt( 1 - quotient )
    else:
        mat = tf.matmul( tf.transpose( tf.conj( X ) ), Y )
        num = tf.abs( tf.trace( mat ) ) ** 2
        dem = int( mat.shape[0] ) ** 2
        return tf.sqrt( 1 - ( num / dem ) )


"""
This module implements the GenericGate Class.

A GenericGate represents a gate with variable location and function.
"""

import tensorflow   as tf
import numpy        as np
import scipy.linalg as la
import itertools    as it

from .pauli import get_pauli_n_qubit_projection
from .pauli import get_pauli_tensor_n_qubit_projection


class GenericGate():
    """
    The GenericGate Class.
    """

    def __init__ ( self, name, num_qubits, gate_size, lm, fun_vals = None,
                   loc_vals = None, parity = None ):
        """
        GenericGate Class Constructor.

        Args:
            name (str): The name of the gate

            num_qubits (int): The number of qubits in the circuit

            gate_size (int): The size of the gate

            lm (LocationModel): The model that maps loc_vals to locations

            fun_vals (List[float]): Initial values for the
                                    gate's function

            loc_vals (List[float]): Initial values for the
                                    gate's location

            parity (int): The side of the topology to occupy. Can be
                          either 0 or 1; prevents consecutive gates
                          from choosing the same location
        """

        if gate_size > num_qubits:
            raise ValueError( "Gate Size must be <= to number of qubits." )

        self.name       = name
        self.num_qubits = num_qubits
        self.gate_size  = gate_size
        self.loc_vals   = loc_vals
        self.fun_vals   = fun_vals
        self.topology   = list( lm.locations ) if parity is None else lm.buckets[ parity ]

        self.num_loc_vars = len( self.topology )
        self.num_fun_vars = 4 ** self.gate_size

        if self.fun_vals is None:
            self.fun_vals = [ np.sqrt( self.num_fun_vars ** -1 ) ] * \
                              self.num_fun_vars

        if self.loc_vals is None:
            self.loc_vals = [0] * self.num_loc_vars

        if len( self.fun_vals ) != self.num_fun_vars:
            raise ValueError( "Incorrect number of function values." )

        if len( self.loc_vals ) != self.num_loc_vars:
            raise ValueError( "Incorrect number of location values." )

        # Construct Tensor
        with tf.variable_scope( self.name ):

            self.fun_vars = [ tf.Variable( val, dtype = tf.float64 )
                              for val in self.fun_vals ]

            self.loc_vars = [ tf.Variable( val, dtype = tf.float64 )
                              for val in self.loc_vals ]

            self.cast_vars = [ tf.cast( x, tf.complex128 )
                               for x in self.fun_vars ]

            gates = []

            for location in self.topology:
                paulis = get_pauli_tensor_n_qubit_projection( self.num_qubits,
                                                              location )

                H = tf.reduce_sum( [ var * pauli for var, pauli
                                     in zip( self.cast_vars, paulis ) ], 0 )

                gates.append( H )

            loc_exps = [ tf.exp( 500 * var ) for var in self.loc_vars ]

            sum_exp = tf.reduce_sum( loc_exps ) + 1e-15

            self.softmax = [ loc_exp / sum_exp for loc_exp in loc_exps ]

            self.cast_max = [ tf.cast( softmax_var, tf.complex128 )
                              for softmax_var in self.softmax ]

            self.herm = tf.reduce_sum( [ softvar * gate
                                         for softvar, gate
                                         in zip( self.cast_max, gates )
                                       ], 0 )

            self.gate = tf.linalg.expm( 1j  * self.herm )

    def get_herm ( self ):
        return self.herm

    def get_gate ( self ):
        return self.gate

    def get_tensor ( self ):
        return self.gate

    def get_fun_vars ( self ):
        return self.fun_vars

    def get_loc_vars ( self ):
        return self.loc_vars

    def get_fun_vals ( self, sess ):
        return sess.run( self.fun_vars )

    def get_loc_vals ( self, sess ):
        return sess.run( self.loc_vars )

    def get_location ( self, sess ):
        loc_idx = np.argmax( sess.run( self.loc_vars ) )
        return self.topology[ loc_idx ]

    def get_unitary ( self, sess ):
        location = self.get_location( sess )
        fun_params = self.get_fun_vals( sess )
        paulis = get_pauli_n_qubit_projection( self.num_qubits, location )
        H = np.sum( [ a*p for a, p in zip( fun_params, paulis ) ], 0 )
        return la.expm( 1j * H )



import argparse
import numpy as np
from qiskit import *


def test_circ_equivalences ( circ1, circ2 ):
    circ1 = circ1.copy()
    circ2 = circ2.copy()
    circ1.remove_final_measurements()
    circ2.remove_final_measurements()
    backend = BasicAer.get_backend( "unitary_simulator" )
    utry1 = qiskit.execute( circ1, backend ).result().get_unitary()
    utry2 = qiskit.execute( circ2, backend ).result().get_unitary()

    if not np.allclose( utry1, utry2 ):
        raise RuntimeError( "Transpiled circuit failed verification." )


if __name__ == "__main__":
    description_info = "Convert QASM to another QASM file containing" + \
                       " only u3 and cx gates."

    parser = argparse.ArgumentParser( description = description_info )

    parser.add_argument( "input_file", type = str,
                         help = "QASM input file" )

    parser.add_argument( "output_file", type = str,
                         help = "Qasm output file" )

    parser.add_argument( "-t", "--test", action = 'store_true',
                         help = "Test output" )

    parser.add_argument( "-m", "--measure", action = 'store_true',
                         help = "Add measurements to output circuit." )

    args = parser.parse_args()

    circ_in = QuantumCircuit.from_qasm_file( args.input_file )
    circ_in.remove_final_measurements()

    circ_out = qiskit.compiler.transpile( circ_in, basis_gates = ['u3', 'cx'] )

    if args.measure:
        circ_out.measure_all()

    if args.test:
        test_circ_equivalences( circ_in, circ_out )

    with open( args.output_file, 'w' ) as f:
        f.write( circ_out.qasm() )



"""
This module implements the functions for the instantiation phase.
"""

import importlib
import pkgutil


import qfast.native as native


discovered_tools = {
    name: importlib.import_module( name )
    for finder, name, ispkg
    in pkgutil.iter_modules( native.__path__, native.__name__ + "." )
}


def list_native_tools():
    """
    List the discovered native tools.

    Returns
        (List[str]): List of discovered tools
    """

    return [ tool.split('.')[-1] for tool in discovered_tools.keys() ]


def get_native_tool ( tool ):
    """
    Retrieves the native tool's module from the discovered list.

    Args:
        tool (str): The native tool

    Return:
        (module): Module associated with the tool specified
    """

    if "qfast.native." + tool not in discovered_tools:
        raise ValueError( "The native tool specified was not found." )

    return discovered_tools[ "qfast.native." + tool ]


def instantiation ( tool, utry ):
    """
    Instantiation uses a native tool to convert a unitary into qasm.

    Args:
        tool (str): The native tool to use

        utry (np.ndarray): The unitary to synthesize

    Return:
        (str): qasm code that implements the unitary
    """

    m = get_native_tool( tool )

    if ( not hasattr( m, "get_native_block_size" ) or
         not hasattr( m, "synthesize" ) ):
        raise TypeError( "The native tool specified has an invalid or "
                         "incomplete api." )

    return m.synthesize( utry )




"""
This module implements the Circuit Class.

A Circuit holds many blocks and implements methods for synthesis.
"""

import os

import numpy as np

from .block import Block
from .decomposition import decomposition, refinement, convert_to_block_list


class Circuit():
    """
    The Circuit Class.
    """

    def __init__ ( self, utry ):
        """
        Circuit Class Constructor.

        Args:
            utry (np.ndarray): Creates a circuit with a single block
        """

        if not isinstance( utry, np.ndarray ):
            raise TypeError( "utry must be a np.ndarray." )

        if len( utry.shape ) != 2:
            raise TypeError( "utry must be a matrix." )

        if utry.shape[0] != utry.shape[1]:
            raise TypeError( "utry must be a square matrix." )

        if not np.allclose( utry @ utry.conj().T, np.identity( len( utry ) ) ):
            raise ValueError( "utry must be a unitary matrix." )

        if not np.allclose( utry.conj().T @ utry, np.identity( len( utry ) ) ):
            raise ValueError( "utry must be a unitary matrix." )

        self.utry = utry
        self.num_qubits = int( np.log2( len( utry ) ) )
        self.blocks = [ Block( self.utry, tuple( range( self.num_qubits ) ) ) ]

    def hierarchically_decompose ( self, native_block_size, **kwargs ):
        """
        Hierarchically decompose a circuit into blocks of size at most
        native_block_size.

        Args:
            native_block_size (int): target block size

        Keyword Args:
            See decomposition in decomposition.py
        """

        # Decompose circuit
        while any( [ block.num_qubits > native_block_size
                     for block in self.blocks ] ):

            new_block_list = []

            for block in self.blocks:

                if block.num_qubits <= native_block_size:
                    new_block_list.append( block )
                else:
                    kwargs["native_block_size"] = native_block_size
                    new_block_list += decomposition( block, **kwargs )

            self.blocks = new_block_list

        # Final Refinement
        fun_vals, loc_fixed = self.get_fun_loc_vals()

        params = { "refinement_distance": 1e-7,
                   "refinement_learning_rate": 1e-6 }

        if "refinement_distance" in kwargs:
            params["refinement_distance"] = \
                kwargs["refinement_distance"]

        if "refinement_learning_rate" in kwargs:
            params["refinement_learning_rate"] = \
                kwargs["refinement_learning_rate"]

        fun_vals = refinement( self.utry, self.num_qubits, len( loc_fixed[0] ),
                               fun_vals, loc_fixed,
                               params["refinement_distance"],
                               params["refinement_learning_rate"] )

        self.blocks = convert_to_block_list( list( range( self.num_qubits ) ),
                                             fun_vals, loc_fixed )

    def get_fun_loc_vals ( self ):
        """
        Converts the circuit's block list into location and function
        value lists.

        Returns:
            (Tuple[List[List[float]], List[Tuple[int]]):
                The circuit block's function values and locations
        """

        return ( [ b.get_fun_vals() for b in self.blocks ],
                 [ b.get_location() for b in self.blocks ] )

    def get_locations ( self ):
        """
        Gets the locations of all the blocks in the circuit.

        Returns:
            (List[Tuple[int]]): The circuit block's locations
        """

        return [ b.get_location() for b in self.blocks ]

    def dump_blocks ( self, directory ):
        """
        Dumps the circuit's blocks into a directory.

        Args:
            directory (str): The directory where blocks will be dumped.
        """

        if not os.path.isdir( directory ):
            raise ValueError( "Invalid directory: %s" % directory )

        for i, block in enumerate( self.blocks ):
            linkname = str( block.get_location() ).replace( ", ", "_" )
            linkname = linkname.replace("(", "").replace(")", "")
            filename = "%d_%s.unitary" % ( i, linkname )
            filename = os.path.join( directory, filename )
            np.savetxt( filename, block.get_utry() )



"""
This module implements the FixedGate Class.

A FixedGate represents a gate with fixed location and variable function.
"""

import tensorflow   as tf
import numpy        as np
import scipy.linalg as la

from .pauli import get_pauli_n_qubit_projection
from .pauli import get_pauli_tensor_n_qubit_projection


class FixedGate():
    """
    The FixedGate Class.
    """

    def __init__ ( self, name, num_qubits, gate_size, loc, fun_vals = None ):
        """
        FixedGate Class Constructor.

        Args:
            name (str): The name of the layer

            num_qubits (int): The number of qubits in the circuit

            gate_size (int): The size of the gate

            loc (Tuple[int]) The qubits this gate acts on (location)

            fun_vals (List[float]): Initial values for the gate's
                                    function
        """

        if gate_size > num_qubits:
            raise ValueError( "Gate Size must be <= to number of qubits." )

        self.name       = name
        self.num_qubits = num_qubits
        self.gate_size  = gate_size
        self.location   = loc
        self.fun_vals   = fun_vals

        self.num_fun_vars = 4 ** self.gate_size

        if self.fun_vals is None:
            self.fun_vals = [ np.sqrt( self.num_fun_vars ** -1 ) ] * \
                              self.num_fun_vars

        if len( self.fun_vals ) != self.num_fun_vars:
            raise ValueError( "Incorrect number of function values." )

        # Construct Tensor
        with tf.variable_scope( self.name ):

            self.fun_vars = [ tf.Variable( val, dtype = tf.float64 )
                              for val in self.fun_vals ]

            self.cast_vars = [ tf.cast( x, tf.complex128 )
                               for x in self.fun_vars ]

            paulis = get_pauli_tensor_n_qubit_projection( self.num_qubits,
                                                          self.location )

            self.herm = tf.reduce_sum( [ var * pauli
                                         for var, pauli
                                         in zip( self.cast_vars, paulis )
                                       ], 0 )

            self.gate = tf.linalg.expm( 1j * self.herm )

    def get_herm ( self ):
        return self.herm

    def get_gate ( self ):
        return self.gate

    def get_tensor ( self ):
        return self.gate

    def get_fun_vars ( self ):
        return self.fun_vars

    def get_fun_vals ( self, sess ):
        return sess.run( self.fun_vars )

    def get_location ( self ):
        return self.location

    def get_unitary ( self, sess ):
        fun_params = self.get_fun_vals( sess )
        paulis = get_pauli_n_qubit_projection( self.num_qubits, self.location )
        H = np.sum( [ a*p for a, p in zip( fun_params, paulis ) ], 0 )
        return la.expm( 1j * H )



"""
This module implements the main decomposition functions.
"""

import logging
import tensorflow as tf
import numpy      as np
import itertools  as it

from .fixedgate import FixedGate
from .genericgate import GenericGate
from .locationmodel import LocationModel
from .block import Block
from .metrics import hilbert_schmidt_distance
from .pauli import get_unitary_from_pauli_coefs, reset_tensor_cache


logger = logging.getLogger( "qfast" )


def decomposition ( block, **kwargs ):
    """
    Decomposes a Block into a sequence of smaller blocks which
    implement the input.

    Args:
        block (Block): Target Block

    Keyword Args:
        start_depth (int): The number of gates to start exploring from

        depth_step (int): The number of added gates in each
                          exploration step

        exploration_distance (float): Exploration's goal distance

        exploration_learning_rate (float): Learning rate of
                                           exploration's optimizer

        refinement_distance (float): Refinement's goal distance

        refinement_learning_rate (float): Learning rate of refinement's
                                          optimizer

        native_block_size (int): Minimum block size

    Returns:
        (List[Block]): Decomposed blocks
    """

    if block.num_qubits <= 2:
        return [ block ]

    params = {}
    params["start_depth" ] = 1
    params["depth_step" ] = 1
    params["exploration_distance" ] = 0.01
    params["exploration_learning_rate" ] = 0.01
    params["refinement_distance" ] = 1e-7
    params["refinement_learning_rate" ] = 1e-6
    params["native_block_size"] = 0
    params.update( kwargs )

    if block.num_qubits <= params["native_block_size"]:
        return [ block ]

    gate_size = get_decomposition_size( block.num_qubits )

    if gate_size < params["native_block_size"]:
        gate_size = params["native_block_size"]

    lm = LocationModel( block.num_qubits, gate_size )

    fun_vals, loc_vals = exploration( block.utry, block.num_qubits, gate_size,
                                      params["start_depth"],
                                      params["depth_step"], lm,
                                      params["exploration_distance"],
                                      params["exploration_learning_rate"] )

    loc_fixed = lm.fix_locations( loc_vals )

    fun_vals = refinement( block.utry, block.num_qubits, gate_size,
                           fun_vals, loc_fixed,
                           params["refinement_distance"],
                           params["refinement_learning_rate"] )

    return convert_to_block_list( block.get_location(), fun_vals, loc_fixed )


def get_decomposition_size ( num_qubits ):
    """
    Given a block size, computes the block size to decompose to

    Args:
        num_qubits (int): Number of qubits in block

    Returns:
        (int): Decomposed block size
    """

    if not isinstance( num_qubits, int ):
        raise TypeError( "Input must be an integer." )

    if num_qubits <= 0:
        raise ValueError( "Number of qubits must be positive." )

    return int( np.ceil( num_qubits / 2 ) )


def exploration ( target, num_qubits, gate_size, start_depth, depth_step,
                  lm, exploration_distance = 0.01, learning_rate = 0.01 ):
    """
    Synthesizes a circuit that implements the target unitary. Explores
    both circuit structure (gate location) and gate functions.

    Args:
        target (np.ndarray): Target unitary

        num_qubits (int): The target unitary's number of qubits

        gate_size (int): number of active qubits in a gate

        start_depth (int): The number of gates to start searching from

        depth_step (int): The number of added gates in each step

        lm (LocationModel): The model that maps loc_vals to locations

        exploration_distance (float): Exploration's goal distance

        learning_rate (float): Learning rate of the optimizer

    Returns:
        (Tuple[List[List[float]], List[List[float]]]):
            The final fun_vals and loc_vals
    """

    depth = start_depth
    fun_vals = [ None ] * depth
    loc_vals = [ None ] * depth
    logger.info( "Starting to search with %d gates." % depth )

    # Stride search over layer_count
    while ( True ):
        result = fixed_depth_exploration( target, num_qubits, gate_size,
                                          fun_vals, loc_vals, lm,
                                          exploration_distance, learning_rate )

        success, fun_vals, loc_vals = result

        if success:
            logger.info( "Found a circuit with %d gates." % depth )
            break

        fun_vals += [ None ] * depth_step
        loc_vals += [ None ] * depth_step
        depth += depth_step
        logger.info( "Added a gate, now at %d gates." % depth )

    # Remember good results
    found_fun_vals = fun_vals
    found_loc_vals = loc_vals

    # Search backwards for a shorter circuit
    for i in range( depth_step - 1 ):
        depth -= 1
        logger.info( "Removing a layer, now at %d gates." % depth )


        result = fixed_depth_exploration( target, num_qubits, gate_size,
                                          fun_vals, loc_vals, lm,
                                          exploration_distance, learning_rate )

        success, fun_vals, loc_vals = result

        if success:
            found_fun_vals = fun_vals
            found_loc_vals = loc_vals
            logger.info( "Found a circuit with %d gates." % depth )

    return found_fun_vals, found_loc_vals


def fixed_depth_exploration ( target, num_qubits, gate_size, fun_vals,
                              loc_vals, lm, exploration_distance = 0.01,
                              learning_rate = 0.01 ):
    """
    Attempts to synthesize the target unitary with a fixed number
    of gates of size gate_size.

    Args:
        target (np.ndarray): Target unitary

        num_qubits (int): The target unitary's number of qubits

        gate_size (int): number of active qubits in a gate

        fun_vals (List[List[float]]): Gate function values

        loc_vals (List[List[float]]): Gate location values

        lm (LocationModel): The model that maps loc_vals to locations

        exploration_distance (float): Exploration's goal distance

        learning_rate (float): Learning rate of the optimizer

    Returns:
        (Tuple[bool, List[List[float]], List[List[float]]]):
            True if succeeded in hitting exploration distance
            and the final fun_vals and loc_vals
    """

    tf.reset_default_graph()
    reset_tensor_cache()

    layers = [ GenericGate( "Gate%d" % i, num_qubits, gate_size, lm,
                            fun_vals[i], loc_vals[i],
                            parity = (i + 1) % lm.num_buckets )
               for i in range( len( fun_vals ) ) ]

    tensor = layers[0].get_tensor()
    for layer in layers[1:]:
        tensor = tf.matmul( layer.get_tensor(), tensor )

    loss_fn   = hilbert_schmidt_distance( target, tensor )
    optimizer = tf.train.AdamOptimizer( learning_rate )
    train_op  = optimizer.minimize( loss_fn )
    init_op   = tf.global_variables_initializer()

    loss_values = []

    with tf.Session() as sess:
        sess.run( init_op )

        while ( True ):
            for i in range( 20 ):
                loss = sess.run( [ train_op, loss_fn ] )[1]

            if loss < exploration_distance:
                logger.info( "Ending exploration at %f distance." % loss )
                return ( True,
                         [ l.get_fun_vals( sess ) for l in layers ],
                         [ l.get_loc_vals( sess ) for l in layers ] )

            loss_values.append( loss )
            logger.debug( "Loss: %f" % loss )

            if len( loss_values ) > 100:
                min_value = np.min( loss_values[-100:] )
                max_value = np.max( loss_values[-100:] )

                # Plateau Detected
                if max_value - min_value < learning_rate / 10:
                    logger.debug( "Ending exploration at %f distance." % loss )
                    return ( False,
                             [ l.get_fun_vals( sess ) for l in layers ],
                             [ l.get_loc_vals( sess ) for l in layers ] )

            if len( loss_values ) > 500:
                min_value = np.min( loss_values[-500:] )
                max_value = np.max( loss_values[-500:] )

                # Plateau Detected
                if max_value - min_value < learning_rate:
                    logger.debug( "Ending exploration at %f distance." % loss )
                    return ( False,
                             [ l.get_fun_vals( sess ) for l in layers ],
                             [ l.get_loc_vals( sess ) for l in layers ] )


def refinement ( target, num_qubits, gate_size, fun_vals, loc_fixed,
                 refinement_distance = 1e-7, learning_rate = 1e-6 ):
    """
    Refines synthesized circuit to better implement the target unitary.
    This is achieved by using fixed circuit structure (gate location)
    and a more fine-grained optimizer.

    Args:
        target (np.ndarray): Target unitary

        num_qubits (int): The target unitary's number of qubits

        gate_size (int): number of active qubits in a gate

        fun_vals (List[List[float]]): Gate function values

        loc_fixed (List[Tuple[int]]): Gate locations

        refinement_distance (float): Refinement's goal distance

        learning_rate (float): Learning rate of the optimizer

    Returns:
        (List[List[float]]): Refined gate function values
    """

    tf.reset_default_graph()
    reset_tensor_cache()

    layers = [ FixedGate( "Gate%d" % i, num_qubits, gate_size,
                          loc = loc_fixed[i], fun_vals = fun_vals[i] )
               for i in range( len( fun_vals ) ) ]

    tensor = layers[0].get_tensor()
    for layer in layers[1:]:
        tensor = tf.matmul( layer.get_tensor(), tensor )

    loss_fn   = hilbert_schmidt_distance( target, tensor )
    optimizer = tf.train.AdamOptimizer( learning_rate )
    train_op  = optimizer.minimize( loss_fn )
    init_op   = tf.global_variables_initializer()

    loss_values = []

    with tf.Session() as sess:
        sess.run( init_op )
        loss = sess.run( loss_fn )
        logger.info( "Starting refinement at %f distance." % loss )

        for i in range( 10000 ):
            for j in range( 50 ):
                loss = sess.run( [ train_op, loss_fn ] )[1]

            if loss < refinement_distance:
                break

            loss_values.append( loss )
            logger.log( 0, loss )

            if len( loss_values ) > 100:
                min_value = np.min( loss_values[-100:] )
                max_value = np.max( loss_values[-100:] )

                # Plateau Detected
                if max_value - min_value < 1e-5:
                    break

            if len( loss_values ) > 500:
                min_value = np.min( loss_values[-500:] )
                max_value = np.max( loss_values[-500:] )

                # Plateau Detected
                if max_value - min_value < 1e-3:
                    break

        logger.info( "Ending refinement at %f distance." % loss )

        return [ l.get_fun_vals( sess ) for l in layers ]


def convert_to_block_list ( block_loc, fun_vals, loc_fixed ):
    """
    Converts the function parameters to unitary matrices and composes
    location into a larger circuit. Returns the resulting block list.

    Args:
        block_loc (Tuple[int]): The location in a larger circuit that
                                this circuit corresponds to.

        fun_vals (List[List[float]]): Gate function values

        locs_fixed (List[Tuple[int]]): Gate locations

    Returns:
        (List[Block]): Resulting block list
    """

    block_list = []

    for loc, fun_params in zip( loc_fixed, fun_vals ):

        # Convert to unitary
        utry = get_unitary_from_pauli_coefs( fun_params )

        # Compose location
        new_loc = tuple( [ block_loc[i] for i in loc ] )

        block_list.append( Block( utry, new_loc ) )

    return block_list
